import sciunit
from sciunit import Capability

class ReceivesEPSCstim(sciunit.Capability):
    """
    Indicates that the model receives an excitatory post-synaptic current (EPSC) shaped input
    """

    def run_EPSCstim(self, dend_loc, weight, tau1, tau2):
        """
        This function must be implemented by the model.

        Must return numpy arrays containing the time and voltage values (at the soma and at the synaptic location )

        Parameters
        ----------
        dend_loc : list
            containing the name of the section (string) and the location (float) where input is received. Eg.: ['dendrite[3]', 0.5]
        weight : float
            weight of the synaptic input
        tau1 : float
            rising time constant of the synaptic input
        tau2 : float
            decay time constant of the synaptic input
        """

        raise NotImplementedError()

    def run_EPSC_stim_get_vm(self, dend_loc, weight, tau1, tau2):
        """
        This function is called by the test and calls the run_EPSCstim() function.
        """

        t, v, v_dend = self.run_EPSCstim(dend_loc, weight, tau1, tau2)

        return t, v, v_dend



from __future__ import division
from builtins import range
from sciunit import Score
import numpy
from sciunit.utils import assert_dimensionless

class ZScore_somaticSpiking(Score):
    """
    Mean of Z scores. A float indicating the sum of standardized difference
    from reference means for somatic spiking features.
    """

    def __init__(self, score, related_data={}):

        if not isinstance(score, Exception) and not isinstance(score, float):
            raise InvalidScoreError("Score must be a float.")
        else:
            super(ZScore_somaticSpiking,self).__init__(score, related_data=related_data)

    @classmethod
    def compute(cls, observation, prediction):
        """Computes average of z-scores from observation and prediction for somatic spiking features"""

        feature_errors=numpy.array([])
        features_names=(list(observation.keys()))
        feature_results_dict={}
        bad_features = []

        for i in range (0, len(features_names)):
            p_value = prediction[features_names[i]]['feature mean']
            o_mean = float(observation[features_names[i]]['Mean'])
            o_std = float(observation[features_names[i]]['Std'])

            p_std = prediction[features_names[i]]['feature sd']


            try:
                feature_error = abs(p_value - o_mean)/o_std
                feature_error = assert_dimensionless(feature_error)
            except ZeroDivisionError:
                feature_error = float("inf")
                feature_error = float("inf")
            except (TypeError,AssertionError) as e:
                feature_error = e
            #feature_errors=numpy.append(feature_errors,feature_error)
            feature_result={features_names[i]: feature_error}

            feature_results_dict.update(feature_result)

            if numpy.isnan(feature_error) or numpy.isinf(feature_error):
                bad_features.append(features_names[i])
            else:
                feature_errors=numpy.append(feature_errors,feature_error)
        score_avg=numpy.nanmean(feature_errors)

        return score_avg, feature_results_dict, features_names, bad_features

    def __str__(self):

        return 'ZScore_avg = %.2f' % self.score




import sciunit
from sciunit import Capability
import multiprocessing


class ProvidesRecordingLocationsOnTrunk(sciunit.Capability):
    """ Indicates that the model provides a list of locations on the trunk (primary apical dendrite) to be tested"""

    def find_trunk_locations(self,distances, tolerance, trunk_origin):
        """
        This function must be implemented by the model.

        Must return two dictionaries

        (1) keys: distances, values: corresponding locations on the trunk (primary apical dendrite) in list
        at 50, 105, 250, 350 um distances from the soma
        The form must be: dend_locations = (dist1, ['trunk_segment1_1',location],['trunk_segment1_2',location]), (dist2, ['trunk_segment2',location]),     (dist3, ['trunk_segment3',location]), (dist4, ['trunk_segment4',location])
        E.g. : OrderedDict([(50, ['dendrite[0]', 0.6956994222486329]), (150, ['dendrite[81]', 0.5557523508251703]), (250, ['dendrite[109]', 0.33250043844278565])])

        (2) keys: locations on the trunk, values: its actual distance from the soma.
        The form must be: actial distances = (('trunk_segment1_1',location), distance), (('trunk_segment2',location), distance),    (['trunk_segment3',location], distance)
        E.g. : {('dendrite[95]', 0.5): 191.4537639215934, ('dendrite[91]', 0.5): 186.10161451767556}

        Parameters
        ----------
        distances : list
            list of dinstances from the soma to be considered. Eg.: [50,150,250,350]
        tolerance : int
            number indicating the range around the distance values in the distances list to be considered. If tolerance = 20, first distance range will be 50+-20 um.
        trunk_origin : list
            first element : name of the section from which the trunk originates, second element : position on section (E.g. ['soma[5]', 1]). If not set by the user, the end of the default soma section is used.      
        """

        raise NotImplementedError()

    def find_trunk_locations_multiproc(self, distances, tolerance, trunk_origin):
        """
        This function is called by the test and calls the find_trunk_locations() function.
        Used to keep all NEURON related tasks in independent processes, to avoid errors like 'template can not be redefined'
        """
        pool_trunk = multiprocessing.Pool(1, maxtasksperchild = 1)
        self.dend_locations, actual_distances = pool_trunk.apply(self.find_trunk_locations, (distances, tolerance, trunk_origin,))  # this way model.dend_loc gets the values
        pool_trunk.terminate()
        pool_trunk.join()
        del pool_trunk
        return self.dend_locations, actual_distances



from builtins import range
import neurom as nm
from neurom.core.dataformat import COLS
from neurom.morphmath import point_dist2
from neurom.viewer import draw
from neuron import h
import neuron
import numpy
import matplotlib.pyplot as plt


""" Authors: Michael Gevaert, Armando Romani, Christian Rossert, Sara Saray"""

def get_apical_point(neurite, morph, tuft_percent=27):
    '''Attempt to find the apical point in 'tufted' neurons
    Consider a neuron:
        |   /    | Tuft = 20%
        |--/     |
        |   /
        |--/
        |
    ----.-----
    All endpoints in the top 'tuft_percent' are found, then their common
    branch segment, furthest from the soma, is identified.
    Args:
        morph: neurom.fst._core.Neuron
        tuft_percent: percentage of the 'height' of the apical dendrite that
        would enclose the tuft, only leaves in this volume are considered as
        endpoints.  Note that this a spherical shell centered at the soma
    Returns:
        Section whose *end* is where the apical branching begins, or None if there
        is a problem
    '''

    apical = neurite
    max_distance2 = -float('inf')

    for leaf in apical.root_node.ileaf():
        point = leaf.points[-1, COLS.XYZ]
        max_distance2 = max(max_distance2,
                            point_dist2(point, morph.soma.center))


    min_distance2 = max_distance2 * (1 - tuft_percent / 100.) ** 2

    common_parents = set(nm.iter_sections(apical))
    #Iterator to the sections in a neurite, neuron or neuron population.
    all_parents = set([])

    for leaf in apical.root_node.ileaf():
        point = leaf.points[-1, COLS.XYZ]
        if min_distance2 <= point_dist2(point, morph.soma.center):
            parents = leaf.iupstream()
            set_parents =  set(parents)
            common_parents &= set_parents
            all_parents |= set_parents

    apical_point_section = None
    for parent_section in nm.iter_sections(apical):
        if parent_section in common_parents:
            common_parents.remove(parent_section)
            if not common_parents:
                #print parent_section
                if parent_section in all_parents:
                    #return parent_section
                    apical_point_section = parent_section
                else:
                    apical_point_section = None

    #return None

    return apical_point_section

def multiple_apical_points(morphology):

    morph = morphology

    apical = [neurite for neurite in morph.neurites
       if nm.NeuriteType.apical_dendrite == neurite.type]


    if not apical:
        L.warning('No apical found')
        return None
    elif 1 < len(apical):
        L.warning('Too many apical dendrites')
        return None

    apical = apical[0]

    dist_apical_point = []
    apical_points_and_distances = []

    closer_apical_points =[]

    point = get_apical_point(apical, morph)

    if point is not None:
        dist = nm.morphmath.point_dist(morph.soma.center, point.points[-1, COLS.XYZ])

        apical_points_and_distances.append([point, dist])


    while any(point[1] < 200 for point in apical_points_and_distances):
        #point = get_apical_point(apical, morph)
        #apical_point.append(point)
        #dist_apical_point.append(nm.morphmath.point_dist(morph.soma.center, point.points[-1, COLS.XYZ]))
        #a, b = min(point[1] for idx, point in enumerate(apical_points_and_distances))
        mn,idx = min( (apical_points_and_distances[i][1],i) for i in range(len(apical_points_and_distances)) )
        #print mn, idx
        #print apical_points_and_distances
        new_apical_points = []
        new_apical_point_added = []

        for child in apical_points_and_distances[idx][0].children:

            root_node = child
            tree = nm.core._neuron.Neurite(root_node)
            point = get_apical_point(tree, morph)
            if point is not None:

                dist = nm.morphmath.point_dist(morph.soma.center, point.points[-1, COLS.XYZ])

                new_apical_point_added.append(True)

                apical_points_and_distances.append([point, dist])
            else:
                new_apical_point_added.append(False)

        """ avoid infinite loop"""
        #print new_apical_point_added
        if any(added == True for added in new_apical_point_added):
            del apical_points_and_distances[idx]
        else:
            closer_apical_points.append(apical_points_and_distances[idx])
            del apical_points_and_distances[idx]

    apical_points_and_distances = apical_points_and_distances + closer_apical_points
    #print apical_points_and_distances

    apical_points = []

    for point in apical_points_and_distances:
        apical_points.append(point[0])

    _, ax = draw(morph)
    for point in apical_points:
        pos = point.points[-1, COLS.XYZ]
        ax.scatter([pos[0], ], [pos[1], ])
    plt.show()

    return apical_points

def get_list_of_diff_section_types(morphology, apical_point_sections):

    morph = morphology

    _, ax = draw(morph)

    apical_trunk_sections = []

    for apical_point_section in apical_point_sections:
        some_trunk_sections = list(apical_point_section.iupstream())
        apical_trunk_sections = list(set(apical_trunk_sections) | set(some_trunk_sections))

    apical_trunk_ids = [section.id for section in apical_trunk_sections]
    apical_trunk_points = [section.points[-1, :2] for section in apical_trunk_sections]
    for point in apical_trunk_points:
            ax.plot(point[0], point[1], 'bo', ms=5)

    apical_tuft_sections = []

    for apical_point_section in apical_point_sections:
        some_tuft_sections = list(apical_point_section.ipreorder())
        #print some_tuft_sections
        apical_tuft_sections = apical_tuft_sections + some_tuft_sections
    """!!!!!!"""
    apical_tuft_sections = [sec for sec in apical_tuft_sections if sec not in apical_point_sections]
    """!!!!!!"""

    apical_tuft_ids = [section.id for section in apical_tuft_sections]
    apical_tuft_points = [section.points[-1, :2] for section in apical_tuft_sections]
    for point in apical_tuft_points:
            ax.plot(point[0], point[1], 'go', ms=5)

    filter1 = lambda n : n.type == nm.APICAL_DENDRITE
    all_sections = list(nm.iter_sections(morph, neurite_filter=filter1))
    all_ids = [section.id for section in all_sections]
    oblique_ids = set(all_ids) - (set(apical_trunk_ids) | set(apical_tuft_ids))
    oblique_sections = [section for section in all_sections if section.id in oblique_ids]
    oblique_points = [section.points[-1, :2] for section in oblique_sections]
    for point in oblique_points:
            ax.plot(point[0], point[1], 'yo', ms=5)

    plt.show()

    section_types = {'tuft' : apical_tuft_sections, 'trunk' : apical_trunk_sections, 'obliques' : oblique_sections}

    return section_types

def get_neuron_isections(icell, list_neurom_sections):
    #icell=h.CCell
    #seclist=icell.apical

    seclist=list(icell.apical)


    xyz = numpy.zeros((len(seclist),3))

    for i, sec in enumerate(seclist):
        #print sec.name()
        iend = int(neuron.h.n3d(sec=sec))-1 # apical point always at endpoint of section
        x = neuron.h.x3d(iend, sec=sec)
        y = neuron.h.y3d(iend, sec=sec)
        z = neuron.h.z3d(iend, sec=sec)
        xyz[i] = [x,y,z]

    neuron_section_list = []

    for section in list_neurom_sections:
        #print section.id

        coord_xyz = section.points[-1, COLS.XYZ]
        #print coord_xyz

        d = (xyz-coord_xyz)**2 # search for minimal distance here
        isec = numpy.argmin(numpy.sum(d, axis=1))
        #print "ISEC", isec

        neuron_section_list.append(isec)

    return neuron_section_list




import sciunit
from sciunit import Capability
import multiprocessing


class ProvidesRandomDendriticLocations(sciunit.Capability):
    """ Indicates that the model provides a list of randomly selected locations on the trunk (primary apical dendrite) to be tested"""

    def get_random_locations(self, num, seed, dist_range, trunk_origin):
        """
        This function must be implemented by the model.

        Must return a list of lists [dendrite, seg.x]. Eg. : [['dendrite[31]', 0.5], ['dendrite[117]', 0.8333333333333333], ['dendrite[117]', 0.16666666666666666], ['dendrite[77]', 0.5], ['dendrite[99]', 0.5]],
        and a dictionary where the keys are the locations, the value is the actual distance of the location from the soma. Eg.: {('dendrite[95]', 0.5): 191.4537639215934, ('dendrite[91]', 0.5): 186.10161451767556}

        Parameters
        ----------
        num : int
            the number of dendritic locations to be selected
        seed : float
            the random seed
        dist_range : list
                containing the mimnimum and maximum distance from the soma. Eg.: [50,150]
        trunk_origin : list
            first element : name of the section from which the trunk originates, second element : position on section (E.g. ['soma[5]', 1]). If not  set by the user, the end of the default soma section is used.
        """

        raise NotImplementedError()

    def get_random_locations_multiproc(self, num, seed, dist_range, trunk_origin):
        """
        This function is called by the test and calls the get_random_locations() function.
        Used to keep all NEURON related tasks in independent processes, to avoid errors like 'template can not be redefined'
        """
        pool = multiprocessing.Pool(1, maxtasksperchild = 1)
        self.dend_locations, actual_distances = pool.apply(self.get_random_locations, (num, seed, dist_range, trunk_origin,))  # this way model.dend_loc gets the values
        pool.terminate()
        pool.join()
        del pool
        return self.dend_locations, actual_distances



from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
#from builtins import str
from builtins import range
from quantities.quantity import Quantity
from quantities import mV, nA
import sciunit
from sciunit import Test,Score
try:
    from sciunit import ObservationError
except:
    from sciunit.errors import ObservationError
import hippounit.capabilities as cap
from sciunit.utils import assert_dimensionless# Converters.
from sciunit.scores import BooleanScore,ZScore # Scores.

try:
    import numpy
except:
    print("NumPy not loaded.")

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

#from neuron import h
import collections
import efel
import os
import multiprocessing
import multiprocessing.pool
import functools
import math
from scipy import stats

import json
from hippounit import plottools
import collections


try:
    import pickle as pickle
except:
    import pickle
import gzip

try:
    import copy_reg
except:
    import copyreg

from types import MethodType

from quantities import mV, nA, ms, V, s

from hippounit import scores

def _pickle_method(method):
    func_name = method.__func__.__name__
    obj = method.__self__
    cls = method.__self__.__class__
    return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
    for cls in cls.mro():
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)


try:
    copy_reg.pickle(MethodType, _pickle_method, _unpickle_method)
except:
    copyreg.pickle(MethodType, _pickle_method, _unpickle_method)


class BackpropagatingAPTest(Test):
    """
    Tests the mode and efficacy of back-propagating action potentials on the apical trunk.

    Parameters
    ----------
    config : dict
        dictionary loaded from a JSON file, containing the parameters of the simulation
    observation : dict
        dictionary loaded from a JSON file, containing the experimental mean and std values for the features to be tested
    force_run : boolean
        If True and the pickle files containing the model's response to the simulation exists, the simulation won't be run again, traces are loaded from the pickle file
    force_run_FindCurrentStim : boolean
        If True and the pickle file containing the adjusted current stimulus parameter exists, the search wont run again, data are loaded from the pickle file
    base_directory : str
        Results will be saved here
    show_plot : boolean
        If False, plots are not displayed but still saved
    save_all : boolean
        If False, only the JSON files containing the absolute feature values, the feature error scores and the final scores, and a log file are saved, but the figures and pickle files are not.
    trunk_origin : list
        first element : name of the section from which the trunk originates, second element : position on section (E.g. ['soma[5]', 1]). If not set by the user, the end of the default soma section is used.
    """

    def __init__(self, config = {},
                observation = {"mean_AP1_amp_at_50um" : None,
                 "std_AP1_amp_at_50um" : None,
                 "mean_AP1_amp_at_150um" : None,
                 "std_AP1_amp_at_150um" : None,
                 "mean_AP1_amp_at_250um" : None,
                 "std_AP1_amp_at_250um" : None,
                 "mean_AP1_amp_strong_propagating_at_350um" : None,
                 "std_AP1_amp_strong_propagating_at_350um" : None,
                 "mean_AP1_amp_weak_propagating_at_350um" : None,
                 "std_AP1_amp_weak_propagating_at_350um" : None,

                 "mean_APlast_amp_at_50um" : None,
                 "std_APlast_amp_at_50um" : None,
                 "mean_APlast_amp_at_150um" : None,
                 "std_APlast_amp_at_150um" : None,
                 "mean_APlast_amp_at_250um" : None,
                 "std_APlast_amp_at_250um" : None,
                 "mean_APlast_amp_at_350um" : None,
                 "std_APlast_amp_at_350um" : None},

                name="Back-propagating action potential test" ,
                force_run=False,
                force_run_FindCurrentStim=False,
                base_directory= None,
                show_plot=True,
                save_all = True,
                trunk_origin = None):

        observation = self.format_data(observation)

        Test.__init__(self, observation, name)

        self.required_capabilities += (cap.ReceivesSquareCurrent_ProvidesResponse_MultipleLocations,
                                        cap.ProvidesRecordingLocationsOnTrunk, cap.ReceivesSquareCurrent_ProvidesResponse,)

        self.force_run = force_run
        self.force_run_FindCurrentStim = force_run_FindCurrentStim

        self.show_plot = show_plot
        self.save_all = save_all

        self.base_directory = base_directory
        self.path_temp_data = None #added later, because model name is needed
        self.path_figs = None
        self.path_results = None
        self.trunk_origin = trunk_origin

        self.logFile = None
        self.test_log_filename = 'test_log.txt'

        self.npool = multiprocessing.cpu_count() - 1

        self.config = config

        description = "Tests the mode and efficacy of back-propagating action potentials on the apical trunk."

    score_type = scores.ZScore_backpropagatingAP

    def format_data(self, observation):
        for key, val in list(observation.items()):
            try:
                assert type(observation[key]) is Quantity
            except Exception as e:
                quantity_parts = val.split(" ")
                number = float(quantity_parts[0])
                units = " ".join(quantity_parts[1:])
                observation[key] = Quantity(number, units)
        return observation

    def spikecount(self, delay, duration, soma_trace):

        trace = {}
        traces=[]
        trace['T'] = soma_trace[0]
        trace['V'] = soma_trace[1]
        trace['stim_start'] = [delay]
        trace['stim_end'] = [delay + duration]
        traces.append(trace)

        traces_results = efel.getFeatureValues(traces, ['Spikecount'])

        spikecount = traces_results[0]['Spikecount']

        return spikecount

    def binsearch(self, model, stim_range, delay, dur, section_stim, loc_stim, section_rec, loc_rec):
        c_minmax = stim_range
        c_step_start = 0.01
        c_step_stop= 0.002

        found = False
        spikecounts = []
        amplitudes = []

        while c_step_start >= c_step_stop and not found:

            c_stim = numpy.arange(c_minmax[0], c_minmax[1], c_step_start)

            first = 0
            last = numpy.size(c_stim, axis=0)-1

            while first <= last and not found:

                midpoint = (first + last)//2
                amplitude = c_stim[midpoint]

                result=[]

                pool = multiprocessing.Pool(1, maxtasksperchild = 1)    # I use multiprocessing to keep every NEURON related task in independent processes

                traces= pool.apply(self.run_cclamp_on_soma, args = (model, amplitude, delay, dur, section_stim, loc_stim, section_rec, loc_rec))
                pool.terminate()
                pool.join()
                del pool

                spikecount = self.spikecount(delay, dur, traces)

                amplitudes.append(amplitude)
                spikecounts.append(spikecount)

                #if spikecount >= 10 and spikecount <=20:
                if spikecount == 15:
                    found = True
                else:
                    #if spikecount > 20:
                    if spikecount > 15:
                        last = midpoint-1
                    #elif spikecount < 10:
                    elif spikecount < 15:
                        first = midpoint+1
            c_step_start=c_step_start/2

        if not found:
            amp_index = min((p for p in range(len(spikecounts)) if spikecounts[p] != 0), key=lambda i: abs(spikecounts[i]-15.0)) # we choose the one that is nearest to 15, but not 0
            # print list(p for p in range(len(spikecounts)) if spikecounts[p] != 0) # this gives the indices where spikecount is not 0, then i takes up these values
            #print amp_index
            amplitude = amplitudes[amp_index]
            spikecount = spikecounts[amp_index]


        binsearch_result=[found, amplitude, spikecount]
        #print binsearch_result

        return binsearch_result

    def run_cclamp_on_soma(self, model, amp, delay, dur, section_stim, loc_stim, section_rec, loc_rec):

        if self.base_directory:
            self.path_temp_data = self.base_directory + 'temp_data/' + 'backpropagating_AP/' + model.name + '/'
        else:
            self.path_temp_data = model.base_directory + 'temp_data/' + 'backpropagating_AP/'


        try:
            if not os.path.exists(self.path_temp_data) and self.save_all:
                os.makedirs(self.path_temp_data)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        file_name = self.path_temp_data + 'soma_traces' + str(amp) + '_nA.p'


        if self.force_run_FindCurrentStim or (os.path.isfile(file_name) is False):
            t, v = model.get_vm(amp, delay, dur, section_stim, loc_stim, section_rec, loc_rec)
            if self.save_all:
                pickle.dump([t, v], gzip.GzipFile(file_name, "wb"))


        else:
            [t, v] = pickle.load(gzip.GzipFile(file_name, "rb"))

        return [t, v]


    def find_current_amp(self, model, delay, dur, section_stim, loc_stim, section_rec, loc_rec):

        print('Finding appropriate current step amplitude...')

        amps = numpy.arange(0.0, 1.1, 0.1)
        #amps= [0.0, 0.2, 0.8]
        #amps=[0.0,0.3,0.8]
        #amps=[0.0,0.2, 0.9]

        pool = multiprocessing.Pool(self.npool, maxtasksperchild=1)


        run_cclamp_on_soma_ = functools.partial(self.run_cclamp_on_soma, model, delay=delay, dur=dur, section_stim=section_stim, loc_stim=loc_stim, section_rec=section_rec, loc_rec=loc_rec)
        traces = pool.map(run_cclamp_on_soma_, amps, chunksize=1)

        pool.terminate()
        pool.join()
        del pool

        spikecounts = []
        _spikecounts = []
        amplitudes = []
        amplitude = None
        spikecount = None

        message_to_logFile = '' # as it can not be open here because then  multiprocessing won't work under python3

        for i in range(len(traces)):
            spikecounts.append(self.spikecount(delay, dur, traces[i]))


        if amps[0] == 0.0 and  spikecounts[0] > 0:

            message_to_logFile += 'Spontaneous firing\n'
            message_to_logFile += "---------------------------------------------------------------------------------------------------\n"

            print('Spontaneous firing')
            amplitude = None
            """TODO: stop the whole thing"""

        elif max(spikecounts) < 10:

            message_to_logFile += 'The model fired at ' + str(max(spikecounts)[0]) + ' Hz to ' + str(amps[-1]) + ' nA current step, and did not reach 10 Hz firing rate as supposed (according to Bianchi et al 2012 Fig. 1 B eg.)\n'
            message_to_logFile += "---------------------------------------------------------------------------------------------------\n"

            print('The model fired at ' + str(max(spikecounts)[0]) + ' Hz to ' + str(amps[-1]) + ' nA current step, and did not reach 10 Hz firing rate as supposed (according to Bianchi et al 2012 Fig. 1 B eg.)')
            amplitude = None

        else:
            for i in range(len(spikecounts)):

                if i != len(spikecounts)-1 and i!= 0:
                    if spikecounts[i] >= 10 and spikecounts[i] <= 20 and (spikecounts[i-1] <= spikecounts[i] and spikecounts[i+1] >= spikecounts[i]):
                        amplitudes.append(amps[i])
                        _spikecounts.append(spikecounts[i])
                    elif spikecounts[i] < 10 and spikecounts[i+1] > 20 and (spikecounts[i-1] <= spikecounts[i] and spikecounts[i+1] >= spikecounts[i]):
                        binsearch_result = self.binsearch(model, [amps[i], amps[i+1]], delay, dur, section_stim, loc_stim, section_rec, loc_rec)
                        amplitude = binsearch_result[1]
                        spikecount = binsearch_result[2]
                elif i==0:  #spikecount[i-1]  is the last element here
                    if spikecounts[i] >= 10 and spikecounts[i] <= 20 and spikecounts[i+1] >= spikecounts[i]:
                        amplitudes.append(amps[i])
                        _spikecounts.append(spikecounts[i])
                    elif spikecounts[i] < 10 and spikecounts[i+1] > 20 and spikecounts[i+1] >= spikecounts[i]:
                        binsearch_result = self.binsearch(model, [amps[i], amps[i+1]], delay, dur, section_stim, loc_stim, section_rec, loc_rec)
                        amplitude = binsearch_result[1]
                        spikecount = binsearch_result[2]
                elif i == len(spikecounts)-1: # there is no spikecounts[i+1] in this case
                    if spikecounts[i] >= 10 and spikecounts[i] <= 20 and spikecounts[i-1] <= spikecounts[i]:
                        amplitudes.append(amps[i])
                        _spikecounts.append(spikecounts[i])
        if len(amplitudes) > 1:
            amp_index = min(list(range(len(_spikecounts))), key=lambda i: abs(_spikecounts[i]-15.0)) # we choose the one that is nearest to 15
            amplitude = amplitudes[amp_index]
            spikecount = _spikecounts[amp_index]

        elif len(amplitudes) == 1:
            amplitude = amplitudes[0]
            spikecount = _spikecounts[0]
        # if len(amplitudes) remained 0, binsearch found an amplitude

        #print amplitude, spikecount

        if spikecount < 10 or spikecount > 20:

            message_to_logFile += 'WARNING: No current amplitude value has been found to which the model\'s firing frequency is between 10 and 20 Hz. The simulation is done using the current amplitude value to which the models fires at a frequency nearest to 15 Hz, but not 0 Hz. Current step amplitude: ' + str(amplitude) + 'nA, frequency: ' + str(spikecount) + 'Hz\n'
            message_to_logFile += "---------------------------------------------------------------------------------------------------\n"

            print('WARNING: No current amplitude value has been found to which the model\'s firing frequency is between 10 and 20 Hz. The simulation is done using the current amplitude value to which the models fires at a frequency nearest to 15 Hz, but not 0 Hz. Current step amplitude: ' + str(amplitude) + 'nA, frequency: ' + str(spikecount) + 'Hz\n')


        if self.base_directory:
            self.path_figs = self.base_directory + 'figs/' + 'backpropagating_AP/' + model.name + '/'
        else:
            self.path_figs = model.base_directory + 'figs/' + 'backpropagating_AP/'


        try:
            if not os.path.exists(self.path_figs) and self.save_all:
                os.makedirs(self.path_figs)
        except OSError as e:
            if e.errno != 17:
                raise
            pass


        plt.figure()
        plt.plot(amps, spikecounts, 'o')
        if amplitude is not None and spikecount is not None:
            plt.plot(amplitude, spikecount, 'o', label = "Chosen current amplitude")
        plt.ylabel('Spikecount')
        plt.xlabel('current amplitude (nA)')
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'Spikecounts_bAP' + '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

        return amplitude, message_to_logFile


    def cclamp(self, model, amp, delay, dur, section_stim, loc_stim, dend_locations):

        if self.base_directory:
            self.path_temp_data = self.base_directory + 'temp_data/' + 'backpropagating_AP/' + model.name + '/'
        else:
            self.path_temp_data = model.base_directory + 'temp_data/' + 'backpropagating_AP/'


        try:
            if not os.path.exists(self.path_temp_data) and self.save_all:
                os.makedirs(self.path_temp_data)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        file_name = self.path_temp_data + 'cclamp_' + str(amp) + '.p'

        traces = {}

        if self.force_run or (os.path.isfile(file_name) is False):
            t, v_stim, v = model.get_multiple_vm(amp, delay, dur, section_stim, loc_stim, dend_locations)

            traces['T'] = t
            traces['v_stim'] = v_stim
            traces['v_rec'] = v #dictionary key: dendritic location, value : corresponding V trace of each recording locations
            if self.save_all:
                pickle.dump(traces, gzip.GzipFile(file_name, "wb"))

        else:
            traces = pickle.load(gzip.GzipFile(file_name, "rb"))

        return traces

    def extract_somatic_spiking_features(self, traces, delay, duration):

        # soma
        trace = {}
        traces_for_efel=[]
        trace['T'] = traces['T']
        trace['V'] = traces['v_stim']
        trace['stim_start'] = [delay]
        trace['stim_end'] = [delay + duration]
        traces_for_efel.append(trace)

        # trunk locations
        '''
        for key in traces['v_rec']:
            for k in traces['v_rec'][key]:
                trace = {}
                trace['T'] = traces['T']
                trace['V'] = traces['v_rec'][key][k]
                trace['stim_start'] = [delay]
                trace['stim_end'] = [delay + duration]
                traces_for_efel.append(trace)
        '''

        efel.setDoubleSetting('interp_step', 0.025)
        efel.setDoubleSetting('DerivativeThreshold', 40.0)

        traces_results = efel.getFeatureValues(traces_for_efel, ['inv_first_ISI','AP_begin_time', 'doublet_ISI'])

        return traces_results

    def extract_amplitudes(self, traces, traces_results, actual_distances):

        #soma_AP_begin_indices = traces_results[0]['AP_begin_indices']

        soma_AP_begin_time = traces_results[0]['AP_begin_time']
        #soma_inv_first_ISI = traces_results[0]['inv_first_ISI']
        soma_first_ISI = traces_results[0]['doublet_ISI'][0]
        #print soma_AP_begin_time[0], soma_AP_begin_time[0]-1
        #print traces_results[0]['inv_first_ISI'], soma_first_ISI
        s_indices_AP1 = numpy.where(traces['T'] >= (soma_AP_begin_time[0]-1.0))
        if 10 < soma_first_ISI:
            plus = 10
        else:
            plus = soma_first_ISI-3
        e_indices_AP1 = numpy.where(traces['T'] >= (soma_AP_begin_time[0]+plus))
        start_index_AP1 = s_indices_AP1[0][0]
        end_index_AP1 = e_indices_AP1[0][0]
        #print start_index_AP1
        #print end_index_AP1

        s_indices_APlast = numpy.where(traces['T'] >= soma_AP_begin_time[-1]-1.0)
        e_indices_APlast = numpy.where(traces['T'] >= soma_AP_begin_time[-1]+10)
        start_index_APlast = s_indices_APlast[0][0]
        end_index_APlast = e_indices_APlast[0][0]

        features = collections.OrderedDict()

        for key, value in traces['v_rec'].items():
            features[key] = collections.OrderedDict()
            for k, v in traces['v_rec'][key].items():
                features[key][k] = collections.OrderedDict()

                features[key][k]['AP1_amp']= float(numpy.amax(traces['v_rec'][key][k][start_index_AP1:end_index_AP1]) - traces['v_rec'][key][k][start_index_AP1])*mV
                features[key][k]['APlast_amp']= float(numpy.amax(traces['v_rec'][key][k][start_index_APlast:end_index_APlast]) - traces['v_rec'][key][k][start_index_APlast])*mV
                features[key][k]['actual_distance'] = actual_distances[k]
        '''
        plt.figure()
        plt.plot(traces['T'],traces['v_stim'], 'r', label = 'soma')
        plt.plot(traces['T'][start_index_AP1],traces['v_stim'][start_index_AP1], 'o', label = 'soma')
        plt.plot(traces['T'][end_index_AP1],traces['v_stim'][end_index_AP1], 'o', label = 'soma')
        plt.plot(traces['T'][start_index_APlast],traces['v_stim'][start_index_APlast], 'o', label = 'soma')
        '''
        # zoom to fist AP
        plt.figure()
        plt.plot(traces['T'],traces['v_stim'], 'r', label = 'soma')
        for key, value in traces['v_rec'].items():
            for k, v in traces['v_rec'][key].items():
                #plt.plot(traces['T'],traces['v_rec'][i], label = dend_locations[i][0]+'('+str(dend_locations[i][1])+') at '+str(self.config['recording']['distances'][i])+' um')
                #plt.plot(traces['T'],traces['v_rec'][key], label = dend_locations[key][0]+'('+str(dend_locations[key][1])+') at '+str(key)+' um')
                plt.plot(traces['T'],traces['v_rec'][key][k], label = k[0]+'('+str(k[1])+') at '+str(actual_distances[k])+' um')

        plt.xlabel('time (ms)')
        plt.ylabel('membrane potential (mV)')
        plt.title('First AP')
        plt.xlim(traces['T'][start_index_AP1], traces['T'][end_index_AP1])
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'AP1_traces'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

        # zom to last AP
        plt.figure()
        plt.plot(traces['T'],traces['v_stim'], 'r', label = 'soma')
        for key, value in traces['v_rec'].items():
            for k, v in traces['v_rec'][key].items():
                #plt.plot(traces['T'],traces['v_rec'][i], label = dend_locations[i][0]+'('+str(dend_locations[i][1])+') at '+str(self.config['recording']['distances'][i])+' um')
                #plt.plot(traces['T'],traces['v_rec'][key], label = dend_locations[key][0]+'('+str(dend_locations[key][1])+') at '+str(key)+' um')
                plt.plot(traces['T'],traces['v_rec'][key][k], label = k[0]+'('+str(k[1])+') at '+str(actual_distances[k])+' um')
        plt.xlabel('time (ms)')
        plt.ylabel('membrane potential (mV)')
        plt.title('Last AP')
        plt.xlim(traces['T'][start_index_APlast], traces['T'][end_index_APlast])
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'APlast_traces'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

        return features


    def plot_traces(self, model, traces, dend_locations, actual_distances):

        if self.base_directory:
            self.path_figs = self.base_directory + 'figs/' + 'backpropagating_AP/' + model.name + '/'
        else:
            self.path_figs = model.base_directory + 'figs/' + 'backpropagating_AP/'


        try:
            if not os.path.exists(self.path_figs) and self.save_all:
                os.makedirs(self.path_figs)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        print("The figures are saved in the directory: ", self.path_figs)

        plt.figure()
        plt.plot(traces['T'],traces['v_stim'], 'r', label = 'soma')
        for key, value in traces['v_rec'].items():
            for k, v in traces['v_rec'][key].items():
                #plt.plot(traces['T'],traces['v_rec'][i], label = dend_locations[i][0]+'('+str(dend_locations[i][1])+') at '+str(self.config['recording']['distances'][i])+' um')
                #plt.plot(traces['T'],traces['v_rec'][key], label = dend_locations[key][0]+'('+str(dend_locations[key][1])+') at '+str(key)+' um')
                plt.plot(traces['T'],traces['v_rec'][key][k], label = k[0]+'('+str(k[1])+') at '+str(actual_distances[k])+' um')

        plt.xlabel('time (ms)')
        plt.ylabel('membrane potential (mV)')
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'traces'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')


    def plot_features(self, model, features, actual_distances):

        observation = self.observation

        model_AP1_amps = numpy.array([])
        model_APlast_amps = numpy.array([])
        exp_mean_AP1_amps_StrongProp = numpy.array([])
        exp_mean_AP1_amps_WeakProp = numpy.array([])
        exp_mean_APlast_amps = numpy.array([])
        exp_std_AP1_amps_StrongProp = numpy.array([])
        exp_std_AP1_amps_WeakProp = numpy.array([])
        exp_std_APlast_amps = numpy.array([])

        distances = []
        dists = numpy.array(self.config['recording']['distances'])
        location_labels = []

        for key, value in features.items():

            if 'mean_AP1_amp_strong_propagating_at_'+str(key)+'um' in list(observation.keys()) or 'mean_AP1_amp_weak_propagating_at_'+str(key)+'um' in list(observation.keys()):
                exp_mean_AP1_amps_StrongProp = numpy.append(exp_mean_AP1_amps_StrongProp, observation['mean_AP1_amp_strong_propagating_at_'+str(key)+'um'])
                exp_std_AP1_amps_StrongProp = numpy.append(exp_std_AP1_amps_StrongProp, observation['std_AP1_amp_strong_propagating_at_'+str(key)+'um'])

                exp_mean_AP1_amps_WeakProp = numpy.append(exp_mean_AP1_amps_WeakProp, observation['mean_AP1_amp_weak_propagating_at_'+str(key)+'um'])
                exp_std_AP1_amps_WeakProp = numpy.append(exp_std_AP1_amps_WeakProp, observation['std_AP1_amp_weak_propagating_at_'+str(key)+'um'])

            else:
                exp_mean_AP1_amps_WeakProp = numpy.append(exp_mean_AP1_amps_WeakProp, observation['mean_AP1_amp_at_'+str(key)+'um'])
                exp_std_AP1_amps_WeakProp = numpy.append(exp_std_AP1_amps_WeakProp, observation['std_AP1_amp_at_'+str(key)+'um'])
                exp_mean_AP1_amps_StrongProp = numpy.append(exp_mean_AP1_amps_StrongProp, observation['mean_AP1_amp_at_'+str(key)+'um'])
                exp_std_AP1_amps_StrongProp = numpy.append(exp_std_AP1_amps_StrongProp, observation['std_AP1_amp_at_'+str(key)+'um'])

            exp_mean_APlast_amps = numpy.append(exp_mean_APlast_amps, observation['mean_APlast_amp_at_'+str(key)+'um'])
            exp_std_APlast_amps = numpy.append(exp_std_APlast_amps, observation['std_APlast_amp_at_'+str(key)+'um'])

            for k, v in features[key].items() :
                distances.append(actual_distances[k])
                model_AP1_amps = numpy.append(model_AP1_amps, features[key][k]['AP1_amp'])
                model_APlast_amps = numpy.append(model_APlast_amps, features[key][k]['APlast_amp'])
                location_labels.append(k[0]+'('+str(k[1])+')')

        plt.figure()
        for i in range(len(distances)):
            plt.plot(distances[i], model_AP1_amps[i], marker ='o', linestyle='none', label = location_labels[i])
        plt.errorbar(dists, exp_mean_AP1_amps_WeakProp, yerr = exp_std_AP1_amps_WeakProp, marker='o', linestyle='none', label = 'experiment - Weak propagating')
        plt.errorbar(dists, exp_mean_AP1_amps_StrongProp, yerr = exp_std_AP1_amps_StrongProp, marker='o', linestyle='none', label = 'experiment - Strong propagating')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('AP1_amp (mV)')
        lgd = plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'AP1_amps'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

        plt.figure()
        for i in range(len(distances)):
            plt.plot(distances[i], model_APlast_amps[i], marker ='o', linestyle='none', label = location_labels[i])
        plt.errorbar(dists, exp_mean_APlast_amps, yerr = exp_std_APlast_amps, marker='o', linestyle='none', label = 'experiment')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('APlast_amp (mV)')
        lgd = plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'APlast_amps'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

    def plot_results(self, observation, prediction, errors, model_name_bAP):


        # Mean absolute feature values plot
        distances = numpy.array(self.config['recording']['distances'])

        model_mean_AP1_amps = numpy.array([])
        model_mean_APlast_amps = numpy.array([])
        model_std_AP1_amps = numpy.array([])
        model_std_APlast_amps = numpy.array([])
        exp_mean_AP1_amps_StrongProp = numpy.array([])
        exp_mean_AP1_amps_WeakProp = numpy.array([])
        exp_mean_APlast_amps = numpy.array([])
        exp_std_AP1_amps_StrongProp = numpy.array([])
        exp_std_AP1_amps_WeakProp = numpy.array([])
        exp_std_APlast_amps = numpy.array([])

        for i in range(len(distances)):
            model_mean_AP1_amps = numpy.append(model_mean_AP1_amps, prediction['model_AP1_amp_at_'+str(distances[i])+'um']['mean'])
            model_mean_APlast_amps = numpy.append(model_mean_APlast_amps, prediction['model_APlast_amp_at_'+str(distances[i])+'um']['mean'])
            model_std_AP1_amps = numpy.append(model_std_AP1_amps, prediction['model_AP1_amp_at_'+str(distances[i])+'um']['std'])
            model_std_APlast_amps = numpy.append(model_std_APlast_amps, prediction['model_APlast_amp_at_'+str(distances[i])+'um']['std'])

            if 'mean_AP1_amp_strong_propagating_at_'+str(distances[i])+'um' in list(observation.keys()) or 'mean_AP1_amp_weak_propagating_at_'+str(distances[i])+'um' in list(observation.keys()):
                exp_mean_AP1_amps_StrongProp = numpy.append(exp_mean_AP1_amps_StrongProp, observation['mean_AP1_amp_strong_propagating_at_'+str(distances[i])+'um'])
                exp_std_AP1_amps_StrongProp = numpy.append(exp_std_AP1_amps_StrongProp, observation['std_AP1_amp_strong_propagating_at_'+str(distances[i])+'um'])

                exp_mean_AP1_amps_WeakProp = numpy.append(exp_mean_AP1_amps_WeakProp, observation['mean_AP1_amp_weak_propagating_at_'+str(distances[i])+'um'])
                exp_std_AP1_amps_WeakProp = numpy.append(exp_std_AP1_amps_WeakProp, observation['std_AP1_amp_weak_propagating_at_'+str(distances[i])+'um'])

            else:
                exp_mean_AP1_amps_WeakProp = numpy.append(exp_mean_AP1_amps_WeakProp, observation['mean_AP1_amp_at_'+str(distances[i])+'um'])
                exp_std_AP1_amps_WeakProp = numpy.append(exp_std_AP1_amps_WeakProp, observation['std_AP1_amp_at_'+str(distances[i])+'um'])
                exp_mean_AP1_amps_StrongProp = numpy.append(exp_mean_AP1_amps_StrongProp, observation['mean_AP1_amp_at_'+str(distances[i])+'um'])
                exp_std_AP1_amps_StrongProp = numpy.append(exp_std_AP1_amps_StrongProp, observation['std_AP1_amp_at_'+str(distances[i])+'um'])

            exp_mean_APlast_amps = numpy.append(exp_mean_APlast_amps, observation['mean_APlast_amp_at_'+str(distances[i])+'um'])
            exp_std_APlast_amps = numpy.append(exp_std_APlast_amps, observation['std_APlast_amp_at_'+str(distances[i])+'um'])

        plt.figure()
        plt.errorbar(distances, model_mean_AP1_amps, yerr = model_std_AP1_amps, marker ='o', linestyle='none', label = model_name_bAP)
        plt.errorbar(distances, exp_mean_AP1_amps_WeakProp, yerr = exp_std_AP1_amps_WeakProp, marker='o', linestyle='none', label = 'experiment - Weak propagating')
        plt.errorbar(distances, exp_mean_AP1_amps_StrongProp, yerr = exp_std_AP1_amps_StrongProp, marker='o', linestyle='none', label = 'experiment - Strong propagating')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('AP1_amp (mV)')
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'AP1_amp_means'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')

        plt.figure()
        plt.errorbar(distances, model_mean_APlast_amps, yerr = model_std_APlast_amps, marker ='o', linestyle='none', label = model_name_bAP)
        plt.errorbar(distances, exp_mean_APlast_amps, yerr = exp_std_APlast_amps, marker='o', linestyle='none', label = 'experiment')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('APlast_amp (mV)')
        lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'APlast_amp_means'+ '.pdf', dpi=600, bbox_extra_artists=(lgd,), bbox_inches='tight')


        # Plot of errors

        keys = []
        values = []

        #fig, ax = plt.subplots()
        plt.figure()
        for key, value in errors.items():
            keys.append(key)
            values.append(value)
        y=list(range(len(keys)))
        y.reverse()
        #ax.set_yticks(y)
        #print keys
        #print values
        plt.plot(values, y, 'o')
        plt.yticks(y, keys)
        if self.save_all:
            plt.savefig(self.path_figs + 'bAP_errors'+ '.pdf', bbox_inches='tight')

    def validate_observation(self, observation):

        for key, value in observation.items():
            try:
                assert type(observation[key]) is Quantity
            except Exception as e:
                raise ObservationError(("Observation must be of the form "
                                        "{'mean':float*mV,'std':float*mV}"))

    def generate_prediction(self, model, verbose=False):
        """Implementation of sciunit.Test.generate_prediction."""

        efel.reset()

        if self.base_directory:
            self.path_results = self.base_directory + 'results/' + 'backpropagating_AP/' + model.name + '/'
        else:
            self.path_results = model.base_directory + 'results/' + 'backpropagating_AP/'

        try:
            if not os.path.exists(self.path_results):
                os.makedirs(self.path_results)
        except OSError as e:
            if e.errno != 17:
                raise
            pass


        global model_name_bAP
        model_name_bAP = model.name

        distances = self.config['recording']['distances']
        tolerance = self.config['recording']['tolerance']

        dend_locations, actual_distances = model.find_trunk_locations_multiproc(distances, tolerance, self.trunk_origin)
        #print dend_locations, actual_distances

        print('Dendritic locations to be tested (with their actual distances):', actual_distances)

        traces={}
        delay = self.config['stimulus']['delay']
        duration = self.config['stimulus']['duration']
        #amplitude = self.config['stimulus']['amplitude']

        prediction = collections.OrderedDict()

        #plt.close('all') #needed to avoid overlapping of saved images when the test is run on multiple models in a for loop
        plt.close('all') #needed to avoid overlapping of saved images when the test is run on multiple models

        amplitude, message_to_logFile = self.find_current_amp(model, delay, duration, "soma", 0.5, "soma", 0.5)

        pool = multiprocessing.Pool(1, maxtasksperchild = 1)
        traces = pool.apply(self.cclamp, args = (model, amplitude, delay, duration, "soma", 0.5, dend_locations))

        filepath = self.path_results + self.test_log_filename
        self.logFile = open(filepath, 'w') # if it is opened before multiprocessing, the multiporeccing won't work under python3

        self.logFile.write('Dendritic locations to be tested (with their actual distances):\n'+ str(actual_distances)+'\n')
        self.logFile.write("---------------------------------------------------------------------------------------------------\n")

        self.logFile.write(message_to_logFile)


        #plt.close('all') #needed to avoid overlapping of saved images when the test is run on multiple models

        self.plot_traces(model, traces, dend_locations, actual_distances)

        traces_results = self.extract_somatic_spiking_features(traces, delay, duration)


        features = self.extract_amplitudes(traces, traces_results, actual_distances)

        features_json = collections.OrderedDict()
        for key in features:
            features_json[key] = collections.OrderedDict()
            for ke in features[key]:
                features_json[key][str(ke)] = collections.OrderedDict()
                for k, value in features[key][ke].items():
                    features_json[key][str(ke)][k] = str(value)


        # generating prediction
        for key in features:
            AP1_amps = numpy.array([])
            APlast_amps = numpy.array([])

            for k in features[key]:
                AP1_amps = numpy.append(AP1_amps, features[key][k]['AP1_amp'] )
            prediction['model_AP1_amp_at_'+str(key)+'um'] = {}
            prediction['model_AP1_amp_at_'+str(key)+'um']['mean'] = float(numpy.mean(AP1_amps))*mV
            prediction['model_AP1_amp_at_'+str(key)+'um']['std'] = float(numpy.std(AP1_amps))*mV

        for key in features:
            AP1_amps = numpy.array([])
            APlast_amps = numpy.array([])
            for k in features[key]:
                APlast_amps = numpy.append(APlast_amps, features[key][k]['APlast_amp'] )
            prediction['model_APlast_amp_at_'+str(key)+'um'] = {}
            prediction['model_APlast_amp_at_'+str(key)+'um']['mean'] = float(numpy.mean(APlast_amps))*mV
            prediction['model_APlast_amp_at_'+str(key)+'um']['std'] = float(numpy.std(APlast_amps))*mV

        prediction_json = collections.OrderedDict()
        for key in prediction:
            prediction_json[key] = collections.OrderedDict()
            for k, value in prediction[key].items():
                prediction_json[key][k]=str(value)


        file_name_json = self.path_results + 'bAP_model_features_means.json'
        json.dump(prediction_json, open(file_name_json, "w"), indent=4)
        file_name_features_json = self.path_results + 'bAP_model_features.json'
        json.dump(features_json, open(file_name_features_json, "w"), indent=4)

        if self.save_all:
            file_name_pickle = self.path_results + 'bAP_model_features.p'

            pickle.dump(features, gzip.GzipFile(file_name_pickle, "wb"))

            file_name_pickle = self.path_results + 'bAP_model_features_means.p'

            pickle.dump(prediction, gzip.GzipFile(file_name_pickle, "wb"))

        self.plot_features(model, features, actual_distances)

        efel.reset()

        return prediction

    def compute_score(self, observation, prediction, verbose=False):
        """Implementation of sciunit.Test.score_prediction."""

        distances = numpy.array(self.config['recording']['distances'])
        #score_sum_StrongProp, score_sum_WeakProp  = scores.ZScore_backpropagatingAP.compute(observation,prediction, [50, 150, 250])
        score_avg, errors= scores.ZScore_backpropagatingAP.compute(observation,prediction, distances)

        scores_dict = {}
        scores_dict['Z_score_avg_strong_propagating'] = score_avg[0]
        scores_dict['Z_score_avg_weak_propagating'] = score_avg[1]

        file_name=self.path_results+'bAP_errors.json'

        json.dump(errors, open(file_name, "w"), indent=4)

        file_name_s=self.path_results+'bAP_scores.json'

        json.dump(scores_dict, open(file_name_s, "w"), indent=4)

        self.plot_results(observation, prediction, errors, model_name_bAP)

        if self.show_plot:
            plt.show()

        if scores.ZScore_backpropagatingAP.strong:#score_avg[0] < score_avg[1]:
            best_score = score_avg[0]
            print('This is a rather STRONG propagating model')


            self.logFile.write('This is a rather STRONG propagating model\n')
            self.logFile.write("---------------------------------------------------------------------------------------------------\n")

            score_json= {'Z_score_avg_STRONG_propagating' : best_score}
        elif scores.ZScore_backpropagatingAP.strong is False:#score_avg[1] < score_avg[0]:
            best_score = score_avg[1]
            print('This is a rather WEAK propagating model')

            self.logFile.write('This is a rather WEAK propagating model\n')
            self.logFile.write("---------------------------------------------------------------------------------------------------\n")

            score_json= {'Z_score_avg_Weak_propagating' : best_score}
        elif scores.ZScore_backpropagatingAP.strong is None:#score_avg[1] == score_avg[0]:
            best_score = score_avg[0]
            score_json= {'Z_score_avg' : best_score}


        file_name_score = self.path_results + 'bAP_final_score.json'
        json.dump(score_json, open(file_name_score, "w"), indent=4)


        score=scores.ZScore_backpropagatingAP(best_score)

        self.logFile.write(str(score)+'\n')
        self.logFile.write("---------------------------------------------------------------------------------------------------\n")


        self.logFile.close()

        self.logFile = self.path_results + self.test_log_filename

        return score

    def bind_score(self, score, model, observation, prediction):

        score.related_data["figures"] = [self.path_figs + 'AP1_amp_means.pdf', self.path_figs + 'AP1_amps.pdf', self.path_figs + 'AP1_traces.pdf',
                                        self.path_figs + 'APlast_amp_means.pdf', self.path_figs + 'APlast_amps.pdf', self.path_figs + 'APlast_traces.pdf', self.path_figs + 'Spikecounts_bAP.pdf',
                                        self.path_figs + 'bAP_errors.pdf', self.path_figs + 'traces.pdf', self.path_results + 'bAP_errors.json',
                                        self.path_results + 'bAP_model_features.json', self.path_results + 'bAP_model_features_means.json',
                                        self.path_results + 'bAP_scores.json', self.path_results + 'bAP_final_score.json', self.path_results + self.test_log_filename]
        score.related_data["results"] = [self.path_results + 'bAP_errors.json', self.path_results + 'bAP_model_features.json', self.path_results + 'bAP_model_features_means.json', self.path_results + 'bAP_scores.json', self.path_results + 'bAP_model_features.p', self.path_results + 'bAP_model_features_means.p', self.path_results + 'bAP_final_score.json']
        return score



from __future__ import division
from sciunit import Score
import numpy
import collections
from sciunit.utils import assert_dimensionless

class ZScore_ObliqueIntegration(Score):
    """
    Average of Z scores. A float indicating the average of standardized difference
    from reference means for oblique integration features.
    """

    def __init__(self, score, related_data={}):

        if not isinstance(score, Exception) and not isinstance(score, float):
            raise InvalidScoreError("Score must be a float.")
        else:
            super(ZScore_ObliqueIntegration,self).__init__(score, related_data=related_data)

    @classmethod
    def compute(cls, observation, prediction):
        """Computes average of z-scores from observation and prediction for features of dendritic integration in oblique dendrites"""
        #print observation
        #print prediction


        errors_dict=collections.OrderedDict()
        errors=[]

        for feat_name, value in observation.items():
            if 'mean' in feat_name:
                p_mean = prediction['model_' + feat_name]
                o_mean = observation[feat_name]
                o_std = observation[feat_name[5:] + '_std']

                try:
                    feature_error = abs(p_mean - o_mean)/o_std
                    feature_error = assert_dimensionless(feature_error)

                except (TypeError,AssertionError) as e:
                    feature_error = e
                errors.append(feature_error)
                errors_dict[feat_name[5:] + '_error'] = feature_error

        #print errors_dict

        score_avg=numpy.nanmean(errors)

        return score_avg, errors_dict

    def __str__(self):

        return 'ZScore_avg = %.2f' % self.score



import sciunit
from sciunit import Capability

class ReceivesMultipleSynapses(sciunit.Capability):
	"""Indicates that the model receives one or multiple synapses"""

	def run_multiple_syn(self, dend_loc, interval, number, weight):
		"""
		This function must be implemented by the model.

		Must return numpy arrays containing the time and voltage values (at the soma and at the synaptic location )

		Parameters
		----------
		dend_loc : list
			containing the name of the section (string) and the location (float) where input is received. Eg.: ['dendrite[3]', 0.5]
		number: int
			number of synaptic input
		interval : float
			time interval between the synaptic inputs
		weight : float
			weight of the synaptic input
		"""

		raise NotImplementedError()

	def run_multiple_synapse_get_vm(self, dend_loc, interval, number, weight):
		"""
		This function is called by the test and calls the run_multiple_syn() function.
		"""
		t, v, v_dend = self.run_multiple_syn(dend_loc, interval, number, weight)

		return t, v, v_dend



#!/usr/bin/env python
"""DocString"""
from __future__ import division

from builtins import range
from past.utils import old_div
from itertools import cycle
import matplotlib
import matplotlib.colors as mplcol
import matplotlib.pyplot as plt
import colorsys
import numpy
import collections


def adjust_spines(ax, spines, color='k', d_out=10, d_down=[]):

    if d_down == []:
        d_down = d_out

    ax.set_frame_on(True)
    ax.patch.set_visible(False)

    for loc, spine in ax.spines.items():
        if loc in spines:
            if loc == 'bottom':
                spine.set_position(('outward', d_down))  # outward by 10 points
            else:
                spine.set_position(('outward', d_out))  # outward by 10 points
            #spine.set_smart_bounds(True)
        else:
            spine.set_visible(False) # set_color('none') # don't draw spine

    # turn off ticks where there is no spine
    if 'left' in spines:
        ax.yaxis.set_ticks_position('left')

        if color is not 'k':

            ax.spines['left'].set_color(color)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color)


    elif 'right' not in spines:
        # no yaxis ticks
        ax.yaxis.set_ticks([])

    if 'right' in spines:
        ax.yaxis.set_ticks_position('right')

        if color is not 'k':

            ax.spines['right'].set_color(color)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color)

    if 'bottom' in spines:
        pass
        ax.xaxis.set_ticks_position('bottom')
        #ax.axes.get_xaxis().set_visible(True)

    else:
        # no xaxis ticks
        ax.xaxis.set_ticks([])
        ax.axes.get_xaxis().set_visible(False)


def light_palette(color, n_colors=6, reverse=False, lumlight=0.8, light=None):

    rgb = mplcol.colorConverter.to_rgb(color)

    if light is not None:
        light = mplcol.colorConverter.to_rgb(light)
    else:
        vals = list(colorsys.rgb_to_hls(*rgb))
        vals[1] = lumlight
        light = colorsys.hls_to_rgb(*vals)

    colors = [color, light] if reverse else [light, color]
    pal = mplcol.LinearSegmentedColormap.from_list("blend", colors)

    palette = pal(numpy.linspace(0, 1, n_colors))

    return palette


def tiled_figure(figname='', frames=1, columns=2, figs=collections.OrderedDict(), axs=None, orientation='page', width_ratios=None, height_ratios=None, top=0.97, bottom=0.05, left=0.07, right=0.97, hspace=0.6, wspace=0.2, dirname=''):

    if figname not in figs:

        if axs == None:
            axs = []

        if orientation == 'landscape':
            figsize=(297/25.4, 210/25.4)
        elif orientation == 'page':
            figsize=(210/25.4, 297/25.4)

        params = {'backend': 'ps',
                  'axes.labelsize': 6,
                  'axes.linewidth' : 0.5,
                  'axes.titlesize': 8,
                  'font.size': 8,
                  'font.size': 8,
                  'axes.titlesize': 8,
                  'legend.fontsize': 8,
                  'xtick.labelsize': 6,
                  'ytick.labelsize': 6,
                  'legend.borderpad': 0.2,
                  'legend.markerscale': 0.1,
                  'legend.loc': 'best',
                  'legend.columnspacing': 4,
                  'text.usetex': False,
                  'pdf.fonttype': 42,
                  'figure.figsize': figsize}
        matplotlib.rcParams.update(params)

        fig = plt.figure(figname, facecolor='white')
        figs[figname] = {}
        figs[figname]['fig'] = fig
        figs[figname]['dirname'] = dirname

        if width_ratios is None:
            width_ratios=[1]*columns

        rows = int(numpy.ceil(frames/float(columns)))
        if height_ratios is None:
            height_ratios=[1]*rows

        gs = matplotlib.gridspec.GridSpec(rows, columns, height_ratios=height_ratios, width_ratios=width_ratios)
        gs.update(top=top, bottom=bottom, left=left, right=right, hspace=hspace, wspace=wspace)

        for fi in range(frames):
            axs.append(fig.add_subplot(gs[int(old_div(fi,columns)), int(fi%columns)]))
            adjust_spines(axs[-1], ['left', 'bottom'], d_out=0)

        figs[figname]['axs'] = axs

    else:
        axs = figs[figname]['axs']

    return axs



from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
#from builtins import str
from builtins import range
from quantities.quantity import Quantity
from quantities import mV, nA
import sciunit
from sciunit import Test,Score
try:
    from sciunit import ObservationError
except:
    from sciunit.errors import ObservationError
import hippounit.capabilities as cap
from sciunit.utils import assert_dimensionless# Converters.
from sciunit.scores import BooleanScore,ZScore # Scores.

try:
    import numpy
except:
    print("NumPy not loaded.")

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

#from neuron import h
import collections
import efel
import os
import multiprocessing
import multiprocessing.pool
import functools
import math
from scipy import stats

import json
from hippounit import plottools
import collections


try:
    import pickle as pickle
except:
    import pickle
import gzip

try:
    import copy_reg
except:
    import copyreg

from types import MethodType

from quantities import mV, nA, ms, V, s

from hippounit import scores

def _pickle_method(method):
    func_name = method.__func__.__name__
    obj = method.__self__
    cls = method.__self__.__class__
    return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
    for cls in cls.mro():
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)


try:
    copy_reg.pickle(MethodType, _pickle_method, _unpickle_method)
except:
    copyreg.pickle(MethodType, _pickle_method, _unpickle_method)


class PSPAttenuationTest(Test):
    """
    Tests how much synaptic potential attenuates from the dendrite (different distances) to the soma.

    Parameters
    ----------
    config : dict
        dictionary loaded from a JSON file, containing the parameters of the simulation
    observation : dict
        dictionary loaded from a JSON file, containing the experimental mean and std values for the features to be tested
    force_run : boolean
        If True and the pickle files containing the model's response to the simulation exists, the simulation won't be run again, traces are loaded from the pickle file
    base_directory : str
        Results will be saved here
    show_plot : boolean
        If False, plots are not displayed but still saved
    save_all : boolean
        If False, only the JSON files containing the absolute feature values, the feature error scores and the final scores, and a log file are saved, but the figures and pickle files are not.
    num_of_dend_locations : int
        Number of randomly selected dendritic locations to be tested
    random_seed: int
        random seed for random dendritic location selection
    trunk_origin : list
        first element : name of the section from which the trunk originates, second element : position on section (E.g. ['soma[5]', 1]). If not set by the user, the end of the default soma section is used.
    """

    def __init__(self, config = {},
                observation = {},

                name="PSP attenuation test" ,
                force_run=False,
                base_directory= None,
                show_plot=True,
                num_of_dend_locations = 15,
                random_seed = 1,
                save_all = True,
                trunk_origin = None):

        observation = self.format_data(observation)

        Test.__init__(self,observation,name)

        self.required_capabilities += (cap.ProvidesRandomDendriticLocations, cap.ReceivesEPSCstim)

        self.force_run = force_run

        self.show_plot = show_plot
        self.save_all = save_all

        self.base_directory = base_directory
        self.path_temp_data = None #added later, because model name is needed
        self.path_figs = None
        self.path_results = None
        self.trunk_origin = trunk_origin

        self.logFile = None
        self.test_log_filename = 'test_log.txt'

        self.npool = multiprocessing.cpu_count() - 1

        self.config = config

        self.num_of_dend_locations = num_of_dend_locations
        self.random_seed = random_seed

        description = "Tests how much synaptic potential attenuates from the dendrite (different distances) to the soma."

    score_type = scores.ZScore_PSPAttenuation

    def format_data(self, observation):

        for key, val in list(observation.items()):
            try:
                observation[key] = float(val)
            except Exception as e:
                quantity_parts = val.split(" ")
                number = float(quantity_parts[0])
                units = " ".join(quantity_parts[1:])
                observation[key] = Quantity(number, units)
        return observation

    def run_stimulus(self, model, locations_weights, tau1, tau2):

        dend, xloc, weight = locations_weights

        traces = {}

        if self.base_directory:
            self.path_temp_data = self.base_directory + 'temp_data/' + 'PSP_attenuation/' + model.name + '/'
        else:
            self.path_temp_data= model.base_directory + 'temp_data/' + 'PSP_attenuation/'

        try:
            if not os.path.exists(self.path_temp_data) and self.save_all:
                os.makedirs(self.path_temp_data)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        file_name = self.path_temp_data + 'stimulus_at_' + dend+ '(' + str(xloc) + ')_weight_' + str(weight) + '.p'

        if self.force_run or (os.path.isfile(file_name) is False):

            print("input at: " + dend + "(" + str(xloc) + ") with weight: " + str(weight))

            t, v, v_dend = model.run_EPSC_stim_get_vm([dend, xloc], weight, tau1, tau2)

            if self.save_all:
                pickle.dump([t, v, v_dend], gzip.GzipFile(file_name, "wb"))

        else:
            t, v, v_dend = pickle.load(gzip.GzipFile(file_name, "rb"))

        traces[dend, xloc] = [t, v, v_dend] # dictionary, the key is the dendritic location
        return traces

    def calculate_weights(self, traces_no_input, EPSC_amp):

        locations_weights = []

        for key, value in traces_no_input.items():
            s = int(len(value[2])*0.9)
            Vm = numpy.mean(value[2][s:])   #calculate mean at the last 10% of the trace, measured at the dendrite
            weight = - EPSC_amp/Vm
            locations_weights.append([key[0], key[1], weight])

        return locations_weights

    def analyse_traces(self, model, traces_dict_no_input, traces_dict, locations_distances):

        if self.base_directory:
            self.path_figs = self.base_directory + 'figs/' + 'PSP_attenuation/' + model.name + '/'
        else:
            self.path_figs = model.base_directory + 'figs/' + 'PSP_attenuation/'


        try:
            if not os.path.exists(self.path_figs) and self.save_all:
                os.makedirs(self.path_figs)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        attenuation_values = {}
        EPSP_amp_values = {}


        '''
        num_of_subplots = len(traces_dict.keys())
        nrows = int(numpy.ceil(numpy.sqrt(num_of_subplots)))
        ncols = int(numpy.ceil(numpy.sqrt(num_of_subplots)))
        i=0
        '''

        dend_depols = {}
        soma_depols = {}


        for key, value in traces_dict.items():

            if not numpy.array_equal(traces_dict[key][0], traces_dict_no_input[key][0]):    #if the  time vectors are not equal, the traces are resampled with fixed time step
                dt = 0.025
                time_vector = numpy.arange(traces_dict[key][0][0], traces_dict[key][0][-1], dt)  #from the first to the last element of the original time vector

                interp_trace_soma = numpy.interp(time_vector, traces_dict[key][0], traces_dict[key][1])
                interp_trace_soma_no_input = numpy.interp(time_vector, traces_dict_no_input[key][0], traces_dict_no_input[key][1])

                interp_trace_dend = numpy.interp(time_vector, traces_dict[key][0], traces_dict[key][2])
                interp_trace_dend_no_input = numpy.interp(time_vector, traces_dict_no_input[key][0], traces_dict_no_input[key][2])

                dend_depol = interp_trace_dend - interp_trace_dend_no_input
                soma_depol = interp_trace_soma - interp_trace_soma_no_input

                dend_depols[key] = dend_depol
                soma_depols[key] = soma_depol

                print("Voltage traces are resampled using linear interpolation")

            else:
                dend_depol = traces_dict[key][2] - traces_dict_no_input[key][2]
                soma_depol = traces_dict[key][1] - traces_dict_no_input[key][1]

                dend_depols[key] = dend_depol
                soma_depols[key] = soma_depol

                time_vector = traces_dict[key][0]

            max_dend_depol = max(dend_depol)
            max_soma_depol = max(soma_depol)
            attenuation = max_soma_depol / max_dend_depol

            attenuation_values[key] = attenuation

            EPSP_amp_values[key] = {'soma' : max_soma_depol, 'dendrite' : max_dend_depol}
            '''
            plt.figure()
            #plt.subplot(nrows, ncols , i+1)
            plt.plot(traces_dict[key][0], soma_depol, label='SOMA')
            plt.plot(traces_dict[key][0], dend_depol, label=key[0]+'('+str(key[1])+')')
            plt.legend(loc = 0)

            #i+=1
            '''
        #sorted_locations_distances = collections.OrderedDict(sorted(locations_distances.items(), key=lambda x: x[1])) # keys are the dendritic locations, values are they distances from soma

        """ Plotting the traces"""
        distances = self.config['target_distances']
        tolerance = self.config['tolerance']

        for dist in distances:

            d = {key:value for (key,value) in list(locations_distances.items()) if value >= dist - tolerance and value <= dist + tolerance }
            sorted_d = collections.OrderedDict(sorted(list(d.items()), key=lambda x: x[1])) # keys are the dendritic locations, values are they distances from soma

            columns = 2
            width_ratios=[1]*columns
            frames = len(list(d.keys()))
            if int(numpy.ceil(frames/float(columns))) < 5:
                rows = 5
            else:
                rows = int(numpy.ceil(frames/float(columns)))
            height_ratios=[1]*rows
            #axs=[]

            fig = plt.figure(figsize = (210/25.4, 297/25.4))
            gs = matplotlib.gridspec.GridSpec(rows, columns, height_ratios=height_ratios, width_ratios=width_ratios)
            gs.update(top=0.92, bottom=0.04, left=0.07, right=0.97, hspace=0.75, wspace=0.3)
            #fig, axes = plt.subplots(nrows=int(round(len(traces_results)/2.0)), ncols=2)
            #fig.tight_layout()
            fig.suptitle('Input at ' + str(dist) + '$\pm$' + str(tolerance) + ' um from soma')
            i=0
            ax = None # needed to be inicialized, because for some distances we may won't have a figure : 'ax' referenced before assignment error
            for key, value in sorted_d.items():
                label_added = False

                #dend_depol = traces_dict[key][2] - traces_dict_no_input[key][2]
                #soma_depol = traces_dict[key][1] - traces_dict_no_input[key][1]

                #plt.subplot(gs[i])
                ax = fig.add_subplot(gs[i])
                if not label_added:
                    plt.plot(time_vector, soma_depols[key]  , label='SOMA')
                    plt.plot(time_vector, dend_depols[key], label='dendrite')
                    label_added = True
                else:
                    plt.plot(time_vector, soma_depols[key])
                    plt.plot(time_vector, dend_depols[key])
                plt.title(key[0]+'('+str("%.2f" % key[1])+') ' + str("%.2f" % locations_distances[key]) + ' um')
                plt.xlabel("ms")
                plt.ylabel("mV")
                i+=1    # next subplot

            #lgd=plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
            if ax:
                handles, labels = ax.get_legend_handles_labels()
                lgd = fig.legend(handles, labels, loc = 'upper right')
            if self.save_all:
                plt.savefig(self.path_figs + 'traces_input_around_' + str(dist)+ '_um' + '.pdf', dpi=600, bbox_inches='tight')

        #print attenuation_values
        return attenuation_values, EPSP_amp_values

    def calcs_and_plots(self, model, attenuation_values, locations_distances, EPSP_amp_values):

        if self.base_directory:
            self.path_figs = self.base_directory + 'figs/' + 'PSP_attenuation/' + model.name + '/'
        else:
            self.path_figs = model.base_directory + 'figs/' + 'PSP_attenuation/'


        try:
            if not os.path.exists(self.path_figs) and self.save_all:
                os.makedirs(self.path_figs)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        print("The figures are saved in the directory: ", self.path_figs)


        distances = self.config['target_distances']
        tolerance = self.config['tolerance']

        observation = self.observation

        obs_means = []
        obs_stds = []

        PSP_attenuation_features = {}
        PSP_attenuation_mean_features = {}
        EPSP_amps = {}

        """ Plot EPSP amplitudes on soma and dendrite"""
        plt.figure()
        i=0 # not to have legend for all the dots
        for key, value in EPSP_amp_values.items():
            EPSP_amps[key] = {'EPSP_amp_soma' : value['soma'], 'EPSP_amp_dendrite' : value['dendrite'], 'distance' : locations_distances[key]}
            if i==0:
                plt.plot(locations_distances[key], value['dendrite'], label = 'dendrite', color= 'black', marker='^', linestyle='none' )
                plt.plot(locations_distances[key], value['soma'], label = 'soma', color= 'black', marker='o', linestyle='none' )
                i += 1
            else:
                plt.plot(locations_distances[key], value['soma'], color= 'black', marker='o', linestyle='none' )
                plt.plot(locations_distances[key], value['dendrite'], color= 'black', marker='^', linestyle='none' )

        plt.xlabel('Synapse distance from soma (um)')
        plt.ylabel('Peak amplitude (mV)')
        plt.title('EPSPs')
        lgd = plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'EPSP_amplitudes'+ '.pdf', dpi=800, bbox_extra_artists=(lgd,), bbox_inches='tight')


        """Plot attenuation values"""

        for dist in distances:
            obs_means.append(observation['mean_attenuation_soma/dend_'+str(dist)+'_um'])
            obs_stds.append(observation['std_attenuation_soma/dend_'+str(dist)+'_um'])
        plt.figure()
        for key, value in attenuation_values.items():
            PSP_attenuation_features[key] = {'attenuation_soma/dendrite' : value, 'distance' : locations_distances[key]}
            plt.plot(locations_distances[key], value, label = key[0]+'('+str(key[1])+') at '+ str(locations_distances[key]) + ' um', marker='o', linestyle='none' )
        plt.errorbar(distances, obs_means, yerr = obs_stds, label = 'experiment', marker='o', linestyle='none', color='r')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('Attenuation soma/dendrite')
        plt.title('PSP attenuation')
        lgd = plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'PSP_attenuation'+ '.pdf', dpi=800, bbox_extra_artists=(lgd,), bbox_inches='tight')


        """ Calculate and plot the mean of attenuation values"""
        label_added = False
        plt.figure()

        for dist in distances:
            att = numpy.array([])
            for key, value in locations_distances.items():
                if value >= dist - tolerance and value < dist + tolerance:
                    att = numpy.append(att, attenuation_values[key])
            mean_att = numpy.mean(att)
            std_att = numpy.std(att)
            PSP_attenuation_mean_features['mean_attenuation_soma/dend_'+str(dist)+'_um']={'mean': float(mean_att), 'std' : float(std_att)}
            if not label_added:
                plt.errorbar(dist, mean_att, yerr = std_att, marker='o', linestyle='none', color = 'b', label=model.name)
                label_added = True
            else:
                plt.errorbar(dist, mean_att, yerr = std_att, marker='o', linestyle='none', color = 'b')
        plt.errorbar(distances, obs_means, yerr = obs_stds, label = 'experiment', marker='o', linestyle='none', color='r')
        plt.xlabel('Distance from soma (um)')
        plt.ylabel('Mean attenuation soma/dendrite')
        plt.title(' Mean PSP attenuation')
        lgd = plt.legend(bbox_to_anchor=(1.0, 1.0), loc = 'upper left')
        if self.save_all:
            plt.savefig(self.path_figs + 'mean_PSP_attenuation'+ '.pdf', dpi=800, bbox_extra_artists=(lgd,), bbox_inches='tight')

        #print PSP_attenuation_features
        return PSP_attenuation_features, PSP_attenuation_mean_features, EPSP_amps


    """ observation contains ratio numbers, have no unit"""
    '''
    def validate_observation(self, observation):

        for key, value in observation.iteritems():
            try:
                assert type(observation[key]) is Quantity
            except Exception as e:
                raise ObservationError(("Observation must be of the form "
                                        "{'mean':float*mV,'std':float*mV}"))
    '''

    def generate_prediction(self, model, verbose=False):
        """Implementation of sciunit.Test.generate_prediction."""

        efel.reset()

        if self.base_directory:
            self.path_results = self.base_directory + 'results/' + 'PSP_attenuation/' + model.name + '/'
        else:
            self.path_results = model.base_directory + 'results/' + 'PSP_attenuation/'

        try:
            if not os.path.exists(self.path_results):
                os.makedirs(self.path_results)
        except OSError as e:
            if e.errno != 17:
                raise
            pass

        distances = self.config['target_distances']
        tolerance = self.config['tolerance']
        dist_range = [min(distances) - tolerance, max(distances) + tolerance]
        tau1 = self.config['tau_rise']
        tau2 = self.config['tau_decay']
        EPSC_amp = self.config['EPSC_amplitude']

        locations, locations_distances = model.get_random_locations_multiproc(self.num_of_dend_locations, self.random_seed, dist_range, self.trunk_origin) # number of random locations , seed
        #print dend_locations, actual_distances
        print('Dendritic locations to be tested (with their actual distances):', locations_distances)

        weight = 0.0

        locations_weights = []
        for locs in locations:
            locs.append(weight)
            locations_weights.append(locs)
        #print locations_weights

        """ run model without an input"""
        pool = multiprocessing.Pool(self.npool, maxtasksperchild=1)
        run_stimulus_ = functools.partial(self.run_stimulus, model, tau1 = tau1, tau2 = tau2)
        traces_no_input = pool.map(run_stimulus_, locations_weights, chunksize=1)

        pool.terminate()
        pool.join()
        del pool
        traces_dict_no_input = dict(list(i.items())[0] for i in traces_no_input) # merge list of dicts into single dict

        locations_weights = self.calculate_weights(traces_dict_no_input, EPSC_amp)

        """run model with inputs"""
        pool = multiprocessing.Pool(self.npool, maxtasksperchild=1)
        run_stimulus_ = functools.partial(self.run_stimulus, model, tau1 = tau1, tau2 = tau2)
        traces = pool.map(run_stimulus_, locations_weights, chunksize=1)

        pool.terminate()
        pool.join()
        del pool
        traces_dict = dict(list(i.items())[0] for i in traces) # merge list of dicts into single dict


        filepath = self.path_results + self.test_log_filename
        self.logFile = open(filepath, 'w') # if it is opened before multiprocessing, the multiporeccing won't work under python3
        self.logFile.write('Dendritic locations to be tested (with their actual distances):\n'+ str(locations_distances)+'\n')
        self.logFile.write("---------------------------------------------------------------------------------------------------\n")


        #plt.close('all') #needed to avoid overlapping of saved images when the test is run on multiple models in a for loop
        plt.close('all') #needed to avoid overlapping of saved images when the test is run on multiple models

        attenuation_values, EPSP_amp_values = self.analyse_traces(model, traces_dict_no_input, traces_dict, locations_distances)
        PSP_attenuation_model_features, PSP_attenuation_mean_model_features, EPSP_amps = self.calcs_and_plots(model, attenuation_values, locations_distances, EPSP_amp_values)

        prediction = PSP_attenuation_mean_model_features

        '''
        for key, value in traces_dict.items():
            plt.figure()
            plt.plot(value[0], value[1], label='SOMA')
            plt.plot(value[0], value[2], label=key[0]+'('+str(key[1])+')')
            plt.legend(loc = 0)
        '''


        PSP_attenuation_model_features_json = {}
        for key, value in PSP_attenuation_model_features.items():
            PSP_attenuation_model_features_json[str(key)]=value

        file_name_features = self.path_results + 'PSP_attenuation_model_features.json'
        json.dump(PSP_attenuation_model_features_json, open(file_name_features, "w"), indent=4)

        EPSP_amps_json = {}
        for key, value in EPSP_amps.items():
            EPSP_amps_json[str(key)]=value

        file_name_EPSP_amps = self.path_results + 'EPSP_amps.json'
        json.dump(EPSP_amps_json, open(file_name_EPSP_amps, "w"), indent=4)

        if self.save_all:
            file_name_features_p = self.path_results + 'PSP_attenuation_model_features.p'
            pickle.dump(PSP_attenuation_model_features, gzip.GzipFile(file_name_features_p, "wb"))

            file_name_EPSP_amps_p = self.path_results + 'EPSP_amps.p'
            pickle.dump(EPSP_amps, gzip.GzipFile(file_name_EPSP_amps_p, "wb"))

        file_name_mean_features = self.path_results + 'PSP_attenuation_mean_model_features.json'
        json.dump(prediction, open(file_name_mean_features, "w"), indent=4)

        efel.reset()

        return prediction

    def compute_score(self, observation, prediction, verbose=False):
        """Implementation of sciunit.Test.score_prediction."""

        distances = self.config['target_distances']

        score_avg, errors= scores.ZScore_PSPAttenuation.compute(observation,prediction, distances)

        file_name=self.path_results+'PSP_attenuation_errors.json'

        json.dump(errors, open(file_name, "w"), indent=4)

        keys = []
        values = []

        plt.figure()
        for key, value in errors.items():
            keys.append(key)
            values.append(value)
        y=list(range(len(keys)))
        y.reverse()
        plt.plot(values, y, 'o')
        plt.yticks(y, keys)
        plt.title('PSP attenuation errors')
        if self.save_all:
            plt.savefig(self.path_figs + 'PSP_attenuation_errors'+ '.pdf', bbox_inches='tight')

        if self.show_plot:
            plt.show()

        score_json= {'score' : score_avg}
        file_name_score = self.path_results + 'PSP_att_final_score.json'
        json.dump(score_json, open(file_name_score, "w"), indent=4)

        score=scores.ZScore_PSPAttenuation(score_avg)

        self.logFile.write(str(score)+'\n')
        self.logFile.write("---------------------------------------------------------------------------------------------------\n")
        self.logFile.close()

        self.logFile = self.path_results + self.test_log_filename

        return score

    def bind_score(self, score, model, observation, prediction):

        score.related_data["figures"] = [self.path_figs + 'PSP_attenuation.pdf', self.path_figs + 'mean_PSP_attenuation.pdf',
                                        self.path_figs + 'PSP_attenuation_errors.pdf', self.path_results + 'PSP_attenuation_model_features.json',
                                        self.path_results + 'PSP_attenuation_mean_model_features.json', self.path_results + 'PSP_attenuation_errors.json',
                                        self.path_results + 'PSP_att_final_score.json', self.path_results + self.test_log_filename]
        score.related_data["results"] = [self.path_results + 'PSP_attenuation_model_features.json', self.path_results + 'PSP_attenuation_mean_model_features.json', self.path_results + 'PSP_attenuation_errors.json', self.path_results + 'PSP_attenuation_model_features.p', self.path_results + 'PSP_att_final_score.json']
        return score



from __future__ import division
from sciunit import Score
import numpy
import math
from sciunit.utils import assert_dimensionless
from quantities import nA

class ZScore_depolblock(Score):
    """
    Z scores. Float indicating standardized difference
    from reference means for current threshold (Ith), and equilibrium potential (Veq) during depolarization block.
    """

    def __init__(self, score, related_data={}):

        #self.score_l=[]
        if not isinstance(score, Exception) and not isinstance(score, float):
            raise InvalidScoreError("Score must be a float.")
        else:
            super(ZScore_depolblock,self).__init__(score, related_data=related_data)

    @classmethod
    def compute(cls, observation, prediction):
        """Computes a z-scores from an observation and a prediction."""

        p_value_I_maxNumAP = prediction['model_I_maxNumAP']
        p_value_I_below_depol_block = prediction['model_I_below_depol_block']
        o_mean_Ith = observation['mean_Ith']
        o_std_Ith = observation['Ith_std']
        p_value_Veq = prediction['model_Veq']
        o_mean_Veq = observation['mean_Veq']
        o_std_Veq = observation['Veq_std']

        try:
            result_I_maxNumAP = abs(p_value_I_maxNumAP - o_mean_Ith)/o_std_Ith
            result_I_maxNumAP = assert_dimensionless(result_I_maxNumAP)
            if not math.isnan(p_value_I_below_depol_block):
                result_I_below_depol_block = abs(p_value_I_below_depol_block - o_mean_Ith)/o_std_Ith
            else:
                result_I_below_depol_block = float('NaN')
            result_I_below_depol_block = assert_dimensionless(result_I_below_depol_block)
            if not math.isnan(p_value_Veq):
                result_Veq = abs(p_value_Veq - o_mean_Veq)/o_std_Veq
            else:
                result_Veq = float('NaN')
            result_Veq = assert_dimensionless(result_Veq)

        except (TypeError,AssertionError) as e:
            result_I_maxNumAP = e
            result_I_below_depol_block = e
            result_Veq = e

        if p_value_I_maxNumAP != p_value_I_below_depol_block and not math.isnan(result_I_below_depol_block): # according to the experiment thesetwo should be equal
            I_diff_penalty = 200.0*(abs(p_value_I_maxNumAP - p_value_I_below_depol_block)/(1*nA))   # divided be (1*nA) to make it dimensionless
            I_diff_penalty = assert_dimensionless(I_diff_penalty)
        else:
            I_diff_penalty = 0

        if math.isnan(result_I_below_depol_block) or math.isnan(result_Veq) :
            final_score = 100.0
        else:
            final_score = numpy.nanmean([result_I_maxNumAP, result_I_below_depol_block, result_Veq]) + I_diff_penalty

        return final_score, result_I_maxNumAP, result_I_below_depol_block, result_Veq, I_diff_penalty

    def __str__(self):

        #return 'Z_Ith = %.2f, Z_Veq = %.2f' % (self.score_l[0], self.score_l[1])
        return 'ZScore_avg+penalty = %.2f' % self.score



"""neurounit: Testing for neuroscience using the sciunit framework."""
import sciunit
import platform
IMPLEMENTATION = platform.python_implementation()
JYTHON = IMPLEMENTATION == 'Jython'
CPYTHON = IMPLEMENTATION == 'CPython'